' A Very Fast Configurable Parser (PB Assembler)
' with Test Demo
' Presented as a FUNCTION for Eros Olmi
'
' Charles E V Pegge
' 8 August 2006
' updated 23 August 2006

' charles@pegge.net
' www.pegge.net
'
' with thanks to Paul Dixon for advising on coding improvements.
' and Michael Mattias for the test material: http://www.talsystems.com/contest12/contest12.zip
'
' The table gives added ease of configuration and flexibility. The upper part of the table
' may be dispensed with if the your source text uses 7 bit ascii only.

' Parsing rules using the default codings table:
'
' space, and all lower ascii and all non alphanumerics are treated as delimiters. eg:
' But an underscore _ (ascii 95) is treated as an alphanumeric eg:
'
'   FREE.LUNCH is interpreted as three words
'   FREE_LUNCH is one word

' all other non alphanumerics including '-' (minus) are self delimiting eg:
'
'   + - . ,  { } are always recognised as individual words even when unspaced
'
'
' But quoted text is treated as an entire word including the outer quotes.
' inner quotes can be embedded eg:
'
'    "She said 'HI' "
'    'She said "HI" '
'
' the bible reading demo will automatically follow if bible12.txt is placed in the
' same folder as this ParseTableTest.exe.
'
'
' quick Ascii code reference:

'  various control chars 0-31:
'  null      0
'  htab      9
'  linefeed  10
'  vtab      11
'  formfeed  12
'  creturn   13
'  escape    27
'  space     32
'  !"#$%&'   33-39
'  ()*+,-./  40-47
'  0-9       48-57
'  :;<=>?@   58-64
'  A-Z       65-90
'  [\]^_`    91-96
'  a-z       97-122
'  {|}~      123-126



#COMPILE EXE
#DIM ALL
#INCLUDE "WIN32API.INC"

DECLARE FUNCTION GetNextToken(BYVAL pString AS BYTE PTR, BYVAL lString AS LONG, BYREF StartPos AS LONG , BYREF NextPos AS LONG) AS LONG

FUNCTION WINMAIN (BYVAL hInstance AS DWORD,BYVAL hPrevInstance AS DWORD,BYVAL lpCmdLine AS ASCIIZ PTR,BYVAL iCmdShow AS LONG) AS LONG

' Lookup Table

DIM tbl AS GLOBAL STRING
 '0123456789ABCDEF'     '   codings
tbl= _                 '
 "             $  " + _ '0  space/ignore      32
 "                " + _ '1  quote marks     ! 33
 " #!####!########" + _ '2  symbols         # 35 min 34
 "0000000000######" + _ '3  numbers         0 48
 "#AAAAAAAAAAAAAAA" + _ '4  alpha uppercase A 65
 "AAAAAAAAAAA####_" + _ '5  alpha uppercase A 65
 "#aaaaaaaaaaaaaaa" + _ '6  alpha lowercase a 97
 "aaaaaaaaaaa#####" + _ '7  alpha lowercase a 97
 "hhhhhhhhhhhhhhhh" + _ '8  HIGHER ASCII    h 104
 "hhhhhhhhhhhhhhhh" + _ '9
 "hhhhhhhhhhhhhhhh" + _ 'A
 "hhhhhhhhhhhhhhhh" + _ 'B
 "hhhhhhhhhhhhhhhh" + _ 'C
 "hhhhhhhhhhhhhhhh" + _ 'D
 "hhhhhhhhhhhhhhhh" + _ 'E
 "hhhhhhhhhhhhhhhh" + _ 'F
 ""
' Coding Table Rules
'
' These 4 rules are hard coded into the Assembler parsing routine:
'
' 1. any code less than 33 is treated as space.
' 2. code 33 signifies a quote, and all characters between and including the matcing quote marks
'    will be included in the word
' 3. any code  from 34 to 47 is treated as self-delimiting symbol (or punctuation).
' 4. any code  greater than 47 is treated as text (so that includes the underscore 95).
'
'
' Configurability
'
' within these rules you can devise your own codings.
' When the GetNextToken is called, it will return the coding associated with the first character of the word.
' It is possible to generate codes for 'return' (13), and 'tab' (9) or any other lower ascii
' if your lexicon requires it. In the above table, a return character (13) will generate a $ (36).
' You can use similar methods for catching Escape codes and other embedded binaries,
' You can designate any character to be a quote marker as well as the usual " (34) and ' (39), by
' placing a ! (33) in the table at the marker's ascii position.

' If you do not want line sensitivity then replace the $ with a space at position 13 in this table.
'
DIM tblp AS GLOBAL BYTE PTR: tblp=STRPTR(tbl)
'
'
DIM dString    AS STRING              ' the text to be read
DIM lString    AS LONG                ' length of this text
DIM pString    AS BYTE PTR            ' actual address of the start of text
DIM TokenStart AS LONG: TokenStart=1  ' string index to the 1st character of a word. (this is an output-only value)
DIM TokenEnd  AS LONG                 ' string index to the position immediately following the word. (set this to wherever you want to begin)
DIM TokenType(127) AS LONG            ' used for counting words by coding-type
'
DIM re AS LONG                        ' the number of repeat reads to be done in the test
'
' sample text
dString="Old_wine    Free.Meal(Lunch,12.34)  "+CHR$(34)+"bon mot (7)"+CHR$(34)+"  "+CHR$(39)+CHR$(34)+"okay"+CHR$(34)+CHR$(39)

re=100000: GOSUB test_and_display ' short sample needs many repeats

' another sample with slighty different lexical rules
ERRCLEAR
OPEN "bible12.txt" FOR INPUT AS 1:CLOSE 1
IF ERR THEN EXIT FUNCTION
OPEN "bible12.txt" FOR BINARY ACCESS READ AS 1:GET$ 1,LOF(1),dString:CLOSE 1

' adaptations for the Bible test:
' double quotes are treated as normal punctuation
' apostrophes (39) are treated as part of word and defined as type 95, like an underscore
' these codings are patched into the table: pointers do it nicely
@tblp[34]=35 ' # ' double quote to be designated as symbol (or punctuation)
@tblp[39]=95 ' _ ' single quote alias apostrophe to be treated like an underscore
'
re=1: GOSUB test_and_display: ' very large text: not too many repeats

EXIT FUNCTION

test_and_display:

LOCAL i AS LONG    ' iterator for repeated reads of the text
LOCAL j AS LONG    ' word counter
LOCAL t AS DWORD   ' for measuring lapsed time in millisecs
LOCAL ot AS DWORD  ' records the lapsed time on a dummy run
                   ' this is used to measure the computing time not spent doing Assembler code

lString=LEN(dString)
pString=STRPTR(dString)  ' address of the string content
                         '
                         ' first a dummy loop to measure the PowerBasic overheads
                         '
DIM s AS STRING          ' string to receive the report to be displayed
DIM dummy AS GLOBAL LONG ' a flag to exclude Assembler code execution
'
'do a dummy run to measure PB and system overheads
'
s=LEFT$(dString,60):dummy=1:GOSUB reading_test
ot=t ' record the overheads time
'
' now do the real thing
s=LEFT$(dString,60)+"..."+$CRLF:dummy=0:GOSUB reading_test
'
'Report
'
MSGBOX s+$CRLF+$CRLF+ _
 STR$(lString)+" bytes"+$CRLF+ _
 STR$(TokenType(48))+" numbers"+$CRLF+ _
 STR$(TokenType(97))+" lowercase words"+$CRLF+ _
 STR$(TokenType(65))+" proper words"+$CRLF+ _
 STR$(TokenType(35))+" punctuation"+$CRLF+ _
 STR$(TokenType(36))+" lines"+$CRLF+ _
 STR$(re*j)+" total items"+$CRLF+ _
 STR$(INT(t-ot))+" milliseconds ASM"+$CRLF+ _
 STR$(INT(ot))+" milliseconds Powerbasic"+$CRLF+ _
 STR$(INT(t))+" milliseconds total time"+$CRLF+ _
 $CRLF+STR$(INT(j/t))+" words/millisec"+$CRLF+ _
 $CRLF+MID$(dString,TokenStart-7,8)+$CRLF+ _
 ""
 '
RETURN
'
reading_test:
DIM ShowWords AS LONG: ShowWords=1
DIM priority AS LONG
FOR i=0 TO 127:tokentype(i)=0:NEXT            ' zero all token counts
j=0                                           ' zero the word counter
priority=GetPriorityClass(GetCurrentProcess)  ' save current thread priority
i=SetPriorityClass(GetCurrentProcess,&h00000100)   ' set Priority to REAL TIME

IF i=0 THEN MSGBOX "unable to get RealTime priority for accurate measurement"

t=getTickCount()                              ' get the current millisecond count since last boot
                                              ' but if you leave your computer on for 49.7 days then
                                              ' this counter will turnover!
FOR i=1 TO re
 TokenStart=0
 TokenEnd=1
 WHILE TokenEnd<=lString
  ' every time you invoke GetNextToken it will get the next word until the end of the string
  ' is reached, then TokenEnd and Token start will be greater than the dString length.
  ' The return value is the token type matching the 1st character of the word.
  ' Here, we use this value to count different token types as they occur.
  ' The counts are reported in the message box above.
  INCR TokenType(GetNextToken(STRPTR(dString), lString, TokenStart, TokenEnd))
  '
  IF ShowWords THEN
   ShowWords=0
   IF LEN(s)<160 THEN
    s=s+MID$(dString,TokenStart,TokenEnd-TokenStart)+$CRLF
    ShowWords=1
   END IF
  END IF
  INCR j
 WEND
NEXT
t=getTickCount()-t 'record lapsed time in milliseconds
SetPriorityClass(GetCurrentProcess(),priority) ' restore priority: this is usually &h00000020 NORMAL PRIORITY
RETURN

END FUNCTION

FUNCTION GetNextToken(BYVAL pString AS BYTE PTR, BYVAL lString AS LONG, BYREF StartPos AS LONG , BYREF NextPos AS LONG) AS LONG
'
IF dummy THEN StartPos=NextPos+1: NextPos=NextPos+6: EXIT FUNCTION
                            '
                            ' first save any registers used for Powerbasic variables
                            ' in this instance, only the ebp register is essential since
                            ' it is used to reference the parameters.
                            ' everything else is already taken care of by the PB Function handling
                            '
! push ebp                  ' save original base pointer
                            '
! mov edx,pString           ' pointer to start of string
! dec edx                   ' offset by -1 for use with index
! mov esi,edx               ' hold this pointer for later use
! mov eax,edx               ' to find end of string
! add edx,lString           ' points to the last character in the string
! mov ebx,NextPos           ' getting parameter BYREF
! add eax,[ebx]             '  points to the character indexed by NextPos
                            ' with basic variables installed,
                            ' it is now safe to overwrite the base pointer register
! mov ebx,tblp              ' move the tbl pointer into ebx ready to do lookups
                            ' we now lose the base pointer until we need it again to assign params
! mov ebp,eax               ' move the char pointer to the base pointer
! mov ecx,eax               ' default value for the start of word pointer
! xor edi,edi               ' clear edi, so it has a zero in it if you go past end of string              '
! jmp short test_space      ' Now the real work
                            '
next_space:                 ' to skip leading spaces tabs etc
                            '
! inc ebp                   ' advance the char pointer
! inc ecx                   ' advance tokenstart pointer
                            '
test_space:                 ' entry point into the next_space loop
                            '
! cmp edx,ebp               ' check if end of string
! jl short past_end         ' exit loop if end of string
! movzx eax, byte ptr [ebp] ' get character
! movzx eax, byte ptr [ebx+eax] ' to look up its type code in tbl
! cmp al,32                 ' is it a space or less ?
! jle short next_space      ' if so then try next character
                            ' drop thru ..
start_of_word:              ' first non space char encountered
                            '
! mov edi,eax               ' save the coding
! cmp al,47                 ' if it is more than a symbol
! jg short next_letter      ' then it is an alphanumeric so try next char
! cmp al,33                 ' is it a quote type of symbol?
! jnz short end_of          ' if not then return a self-delimiting symbol
                            ' drop thru for quotes
quote_string:               ' this is a quoted item, we need to find the end quote mark
                            '
! mov al,[ebp]              ' reload the specific quote mark
                            '
quote_string_traverse:      ' look for a matching quote mark

! inc ebp                   ' advance pointer for next char
! cmp edx,ebp               ' check if past the end of string
! jl short end_of           ' finish if so
! cmp al,[ebp]              ' check for matching quote mark
! jnz short quote_string_traverse
                            ' if no match then try next
! jmp short end_of          ' we have our quote so finish
                            '
next_letter:                ' scanning an alpha numeric. Could be delimited by a symbol, space or lesser ascii
                            '
! inc ebp                   ' advance pointer
! cmp edx,ebp               ' past end of string?
! jl short end_of           ' if so then finish
! movzx eax,byte ptr [ebp]  ' get char
! mov al,[ebx+eax]          ' then lookup its type in tbl
! cmp al,47                 ' if it is more than a space or symbol
! jg short next_letter      ' then it is an alphanumeric so try then next char
! jmp short past_end        '
                            '
end_of:                     ' results are ready to pass back to BASIC variables
                            '
! inc ebp                   ' increment char pointer
                            '
past_end:                   ' jump here when beyond end of string
                            '
! mov eax,ebp               ' hold the end position in eax before we pop the bp register
! sub ecx,esi               ' derive our start of word index
! sub eax,esi               ' derive the after-word index
                            '
                            '
                            ' it is now safe to restore Powerbasic's reserved registers
! pop ebp                   ' restore base pointer register
                            ' we can now return results BYREF
! mov ebx,StartPos          '
! mov [ebx],ecx             '
! mov [ebx-4],eax           ' NextPos is next parameter along
                            ' so it is now safe to store to any PowerBASIC variables
! mov FUNCTION,edi          ' store coding token type
                            '
                            ' if the string is null or contains no words or symbols
                            ' then StartPos will be greater than the length of the string

END FUNCTION
